﻿using System;

namespace Plux {
  public class ScriptCompletedEventArgs : EventArgs {
    private readonly IScriptResult result;

    public IScriptResult Result { get { return result; } }

    public ScriptCompletedEventArgs(IScriptResult result)
        : base() {
      this.result = result;
    }
  }


  // represents a plain table structure
  // invariant: forall 0 <= x < Entries.Count : Captions.Count == Entries[x].Count
  public interface IScriptTable {
    string[] Captions { get; }
    // entries in form of [Row][Column]
    // every row has the same number of columns
    // entries might be null
    // the same column might contain different types of objects in different rows
    object[][] Entries { get; }
  }


  [SlotDefinition("ScriptProvider")]
  // Default: 100, lower value = higher priority
  [Param("Priority", typeof(byte))]
  public interface IScriptProvider {
      // may contain function/alias definitions only
      string ScriptBody { get; }
  }

  [SlotDefinition("ScriptExecutor")]
  public interface IScriptExecutor {
    IScriptSession CreateSession(ExtensionInfo owner);

    // methods for script execution without explicit sessions
    // the same session may be used for different extensions
    // variables are not guaranteed to persist

    // use null for EventHandler to discard result notification
    void ExecuteAsync(string command, EventHandler<ScriptCompletedEventArgs> onCompletedEventHandler);
    IScriptResult ExecuteSync(string command);

    // use null for EventHandler to discard result notification
    void ExecutePriorAsync(string command, EventHandler<ScriptCompletedEventArgs> onCompletedEventHandler);

    // May be called iterative with previous results: e.g. g -> get- -> get- {extensions,slots, ...}
    // input: the statement
    // pos: position of cursor (to determine part of statement)
    // cut_start: remove from this position
    // cut_end: to this position
    // and replace with one string from return value
    /* test it in the _old_ console with the "complete" command, use % to simulate cursor position
         * e.g. complete get%
         *      complete get-%
         *      complete get-ext% | register-
         */
    string[] CompleteCommand(string input, int pos, out int cut_start, out int cut_end);
  }


  public interface IScriptSession {
    event EventHandler<ScriptCompletedEventArgs> ScriptCompleted;

    IScriptResult Result { get; }

    object GetVariable(string variableName);
    void SetVariable(string variableName, object value);
    // An init script which only applies to this session
    string InitScript { get; set; }

    void Close();

    void ExecuteAsync(string command);
    void ExecutePriorAsync(string command);
    void ExecuteSync(string command);
  }


  public interface IScriptError {
    Exception Exception { get; }
    object TargetObject { get; }
  }


  public interface IScriptMessage {
      // current tags: {debug,verbose,commanddetail,warning,progress}
    string Tag { get; }
    string Text { get; }
  }


  public interface IScriptResult {
    string Command { get; }

    IScriptError[] Errors { get; }
    IScriptMessage[] Messages { get; }
    IScriptResult[] SubscriptResults { get; }
    object[] Output { get; }
    CommandState State { get; }
    string StateMessage { get; }
  }


  public interface IScriptAction {
    // the command to execute
    string Command { get; }
    // description of the action
    string Description { get; }
    // indicates that the action should be executed immediately
    bool ExecuteImmediate { get; }
    // the cursor position for further editing (if not executed immediately)
    int CursorPosition { get; }
  }


  [SlotDefinition("RelatedScriptActions")]
  public interface IRelatedScriptActions {
    Type[] SupportedTypes { get; }
    // varname: Name of the variable the object will be bound to (e.g. "obj" ($obj))
    IScriptAction[] GetActions(object obj, string varname);
  }


  [SlotDefinition("Plux.ScriptRunner")]
  public interface IScriptRunner {
    void ExecuteStartupScripts();
    void ExecuteScriptFile(string path);
    void ExecuteScriptResource(PluginInfo pluginInfo, string scriptName);
    void ExecuteScript(string script);
  }
}