using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;

namespace Plux.Framework {
  [Extension(OnCreated = "OnCreated")]
  [Plug("Analyser")]
  [ParamValue("SupportedExtensions", ".dll;.exe")]
  public class AssemblyAnalyser : IAnalyser {
    private ExtensionInfo ext;
    readonly LogEventHandler Log;

    public AssemblyAnalyser() {
      Log = OnLogEvent;
    }

    public AssemblyAnalyser(LogEventHandler logHandler) {
      Log = logHandler;
    }

    public void OnCreated(object sender, ExtensionEventArgs args) {
      ext = args.ExtensionInfo;
    }

    public PluginInfo AnalysePlugin(FileInfo fileInfo, PluginBuilder builder) {
      if(!fileInfo.Extension.Equals(".dll", StringComparison.OrdinalIgnoreCase))
        return null;
      Assembly assembly = GetAssembly(fileInfo);
      if (!IsAssemblyPlugin(assembly)) return null;

      Log(this, new LogEventArgs(
            MessageType.Info, VerbosityLevel.Detailed,
            Messages.PluginDiscovered, fileInfo.Name,
            String.Format("\n  ({0})", fileInfo.FullName)));

      List<ExtensionTypeInfo> extensionTypes = new List<ExtensionTypeInfo>();
      foreach (Type type in assembly.GetTypes()) {
        List<ExtensionAttribute> extensionAttributes = GetCustomAttributes<ExtensionAttribute>(type);
        if (extensionAttributes.Count == 0) continue;

        List<PlugAttribute> plugAttributes = GetCustomAttributes<PlugAttribute>(type);
        List<SlotAttribute> slotAttributes = GetCustomAttributes<SlotAttribute>(type);
        List<ParamValueAttribute> paramValueAttributes = GetCustomAttributes<ParamValueAttribute>(type);
        List<MonitorAttribute> monitorAttributes = GetCustomAttributes<MonitorAttribute>(type);

        foreach (ExtensionAttribute extensionAttribute in extensionAttributes) {
          Log(this, new LogEventArgs(
                      MessageType.Info, VerbosityLevel.Diagnostic,
                      Messages.ExtensionFound, extensionAttribute.Name));
          List<PlugTypeInfo> plugTypes = new List<PlugTypeInfo>();
          foreach (PlugAttribute plugAttribute in plugAttributes) {
            Log(this, new LogEventArgs(
                        MessageType.Info, VerbosityLevel.Diagnostic,
                        Messages.PlugFound, plugAttribute.ForSlot));
            List<ParamValue> paramValues = new List<ParamValue>();
            foreach (ParamValueAttribute paramValueAttribute in paramValueAttributes) {
              if (paramValueAttribute.Plug != String.Empty && !paramValueAttribute.Plug.Equals(plugAttribute.ForSlot))
                continue;
              Log(this, new LogEventArgs(
                          MessageType.Info, VerbosityLevel.Diagnostic,
                          Messages.ParamValueFound, paramValueAttribute.Name));
              paramValues.Add(builder.CreateParamValue(paramValueAttribute));
            }
            plugTypes.Add(builder.CreatePlugType(plugAttribute, paramValues));
          }
          List<SlotTypeInfo> slotTypes = new List<SlotTypeInfo>();
          foreach (SlotAttribute slotAttribute in slotAttributes) {
            Log(this, new LogEventArgs(
                        MessageType.Info, VerbosityLevel.Diagnostic,
                        Messages.SlotFound, slotAttribute.Name));
            slotTypes.Add(builder.CreateSlotType(slotAttribute));
          }
          List<MonitorInfo> monitors = new List<MonitorInfo>();
          foreach (MonitorAttribute monitorAttribute in monitorAttributes) {
            monitors.Add(builder.CreateMonitor(monitorAttribute));
          }
          extensionTypes.Add(builder.CreateExtensionType(extensionAttribute, type, plugTypes, slotTypes, monitors));
        }
      }
      return builder.CreatePlugin(fileInfo, assembly, extensionTypes);
    }

    public ContractInfo AnalyseContract(FileInfo fileInfo, PluginBuilder builder) {
      if(!fileInfo.Extension.Equals(".dll", StringComparison.OrdinalIgnoreCase))
        return null;
      Assembly assembly = GetAssembly(fileInfo);
      if (!IsAssemblyContract(assembly)) return null;

      Log(this, new LogEventArgs(
                  MessageType.Info, VerbosityLevel.Detailed,
                  Messages.ContractDiscovered, fileInfo.Name,
                  String.Format("\n  ({0})", fileInfo.FullName)));
      List<SlotDefinition> slotDefinitions = new List<SlotDefinition>();
      foreach (Type type in assembly.GetTypes()) {
        List<SlotDefinitionAttribute> slotDefinitionAttributes = GetCustomAttributes<SlotDefinitionAttribute>(type);
        if (slotDefinitionAttributes.Count == 0) continue;

        List<ParamAttribute> paramAttributes = GetCustomAttributes<ParamAttribute>(type);

        foreach (SlotDefinitionAttribute slotDefinitionAttribute in slotDefinitionAttributes) {
          Log(this, new LogEventArgs(
                      MessageType.Info, VerbosityLevel.Diagnostic,
                      Messages.SlotDefinitionFound, slotDefinitionAttribute.Name));
          List<ParamDefinition> paramDefinitions = new List<ParamDefinition>();
          foreach (ParamAttribute paramAttribute in paramAttributes) {
            Log(this, new LogEventArgs(
                        MessageType.Info, VerbosityLevel.Diagnostic,
                        Messages.ParamDefinitionFound, paramAttribute.Name));
            paramDefinitions.Add(builder.CreateParamDefinition(paramAttribute));
          }
          slotDefinitions.Add(builder.CreateSlotDefinition(slotDefinitionAttribute, type, paramDefinitions));
        }
      }
      return builder.CreateContract(fileInfo, assembly, slotDefinitions);
    }

    void OnLogEvent(object sender, LogEventArgs args) {
      ext.LogEvent(args);
    }

    public static bool IsAssemblyPlugin(Assembly assembly) {
      if (assembly != null) {
        foreach (Type type in assembly.GetTypes())
          if (HasCustomAttribute<ExtensionAttribute>(type))
            return true;
      }
      return false;
    }

    public static bool IsAssemblyContract(Assembly assembly) {
      if (assembly != null) {
        foreach (Type type in assembly.GetTypes())
          if (HasCustomAttribute<SlotDefinitionAttribute>(type))
            return true;
      }
      return false;
    }

    public static Assembly GetAssembly(FileInfo file) {
      Assembly assembly = null;
      assembly = Assembly.ReflectionOnlyLoadFrom(file.FullName);

      AppDomain.CurrentDomain.ReflectionOnlyAssemblyResolve +=
        OnReflectionOnlyAssemblyResolve;
      return assembly;
    }

    internal static Assembly OnReflectionOnlyAssemblyResolve(object sender, ResolveEventArgs args) {
      return Assembly.ReflectionOnlyLoad(args.Name);
    }

    private static bool HasCustomAttribute<T>(MemberInfo type) {
      foreach (CustomAttributeData cad in
        CustomAttributeData.GetCustomAttributes(type)) {
        if (cad.Constructor.DeclaringType.AssemblyQualifiedName
            == typeof(T).AssemblyQualifiedName)
          return true;
      }
      return false;
    }

    public static List<T> GetCustomAttributes<T>(Type type) where T : Attribute {
      List<T> list = new List<T>();
      if (type.Assembly.ReflectionOnly) {
        foreach (CustomAttributeData cad in CustomAttributeData.GetCustomAttributes(type)) {
          Type cadType = cad.Constructor.DeclaringType;
          //Type compType = typeof(T);
          string compTypeName = typeof(T).AssemblyQualifiedName;
          while (cadType != null) {
            if (compTypeName == cadType.AssemblyQualifiedName) {
              list.Add(DeserializeAttribute<T>(cad));
              break;
            }
            cadType = cadType.BaseType;
          }

          //if(cad.Constructor.DeclaringType.AssemblyQualifiedName
          //   == typeof(T).AssemblyQualifiedName) {
          //  list.Add(DeserializeAttribute<T>(cad));
          //}
        }
      } else {
        Attribute[] attributes = Attribute.GetCustomAttributes(type, typeof(T), true);
        list.AddRange((T[])attributes);
      }
      return list;
    }

    private static T DeserializeAttribute<T>(CustomAttributeData cad) {
      Type[] types = new Type[cad.ConstructorArguments.Count];
      object[] args = new object[cad.ConstructorArguments.Count];
      for (int i = 0; i < args.Length; ++i) {
        CustomAttributeTypedArgument arg = cad.ConstructorArguments[i];
        types[i] = Type.GetType(arg.ArgumentType.AssemblyQualifiedName);
        args[i] = GetValue(arg);
      }

      Type attrType = Type.GetType(cad.Constructor.DeclaringType.AssemblyQualifiedName);
      ConstructorInfo constr = attrType.GetConstructor(types);
      T attr = (T)constr.Invoke(args);

      foreach (CustomAttributeNamedArgument arg in cad.NamedArguments) {
        switch (arg.MemberInfo.MemberType) {
          case MemberTypes.Field: {
              FieldInfo info = attrType.GetField(arg.MemberInfo.Name);
              info.SetValue(attr, GetValue(arg.TypedValue));
            }
            break;
          case MemberTypes.Property: {
              PropertyInfo info = attrType.GetProperty(arg.MemberInfo.Name);
              info.SetValue(attr, GetValue(arg.TypedValue), null);
            }
            break;
        }
      }

      return attr;
    }


    private static object GetValue(CustomAttributeTypedArgument arg) {
      Type actual = arg.Value.GetType(), expected = arg.ArgumentType;

      //if (actual == expected)
      //  return arg.Value;
      if (expected.IsAssignableFrom(actual))
        return arg.Value;

      Type real = Type.GetType(expected.AssemblyQualifiedName);

      if (expected.IsArray) {
        // arg.Value is some crazy System.Collections.ObjectModel.RepositoryCollection
        // -> Convert it to the expected array-type.

        if (expected.GetArrayRank() != 1) {
          throw new ArrayTypeMismatchException(
            "Only one-dimensional arrays are supported!");
        }

        IList src = arg.Value as IList;
        if (src != null) {
          int elemCount = src.Count;

          ConstructorInfo info = real.GetConstructor(new Type[] { typeof(int) });
          object arr = info.Invoke(new object[] { elemCount });

          IList dest = arr as IList;
          for (int i = 0; i < elemCount; ++i)
            if (dest != null) dest[i] = GetValue((CustomAttributeTypedArgument)src[i]);

          return arr;
        }
      }

      if (expected.IsEnum) {
        // arg.Value is an Int32 -> Convert it to the expected enum-type.
        return Enum.Parse(real, ((Int32)arg.Value).ToString());
      }

      throw new Exception("Couldn't get value of " + arg);
    }

  }
}