﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using Plux;
using Plux.Framework;


internal static class QuickDialog {
  internal static void Register(Form parent) {
    var plugs = new List<PlugTypeInfo>();
    foreach (var plug in Runtime.Repository.PlugTypeInfos) {
      if (plug.RegisteredInSlots.Count > 0) {
        continue;
      }
      plugs.Add(plug);
    }
    object sel = ChooseDialog(parent, plugs, "Register Extension", FormatRegister);
    if (sel != null) {
      ((PlugTypeInfo) sel).Register();
    }
  }

  internal static void Deregister(Form parent) {
    List<PlugTypeInfo> plugs = GetRegisteredPlugs();
    object sel = ChooseDialog(parent, plugs, "Deregister Extension", FormatDeregister);
    if (sel != null) {
      ((PlugTypeInfo) sel).Deregister();
    }
  }

  internal static void Plug(Form parent) {
    var plugs = GetRegisteredPlugs();
    object sel = ChooseDialog(parent, plugs, "Plug Extension", FormatPlug);
    if (sel != null) {
      var plug = (PlugTypeInfo) sel;
      foreach (var slot in plug.RegisteredInSlots) {
        plug.Plug(slot);
      }
    }
  }

  internal static void Unplug(Form parent) {
    var plugs = GetPluggedPlugs();
    object sel = ChooseDialog(parent, plugs, "Unplug Extension", FormatUnplug);
    if (sel != null) {
      var plug = (PlugInfo) sel;
      plug.Unplug();
    }
  }

  internal static void Select(Form parent) {
    var plugs = GetPluggedPlugs();
    object sel = ChooseDialog(parent, plugs, "Select Extension", FormatUnplug);
    if (sel != null) {
      var plug = (PlugInfo) sel;
      var slot = plug.PluggedInSlots[0];
      slot.SelectedPlug = plug;
    }
  }

  internal static void Restore(Form parent) {
    List<CompositionRestorer.SavedCompilationInfo> storedComps = GetSavedCompilations();
    storedComps.Sort();
    object sel = RestoreDialog(parent, storedComps, "Restore Composition", FormatRestore);
    if (sel != null) {
      var storedComp = (CompositionRestorer.SavedCompilationInfo)sel;
      foreach (SlotInfo si in storedComp.RootSlots) {
        si.Deregister();
      }
      Runtime.TaskQueue.InvokeOnEmptied(CompleteRestore, 
        new object[] {storedComp.StoredFile, storedComp.Root});
    }
  }

  static void CompleteRestore(object args) {
    object[] argsArray = (object[]) args;
    CompositionRestorer.Restore((string) argsArray[0], (ExtensionInfo) argsArray[1]);
  }

  internal static void Save(Form parent) {
    var slots = Runtime.Repository.SlotInfos;
    var sortedSlots = new List<SlotInfo>(slots);
    sortedSlots.Sort(CompareSlotInfoAlphabetical);
    object sel = SaveDialog(parent, sortedSlots, "Save Composition", FormatSlot);
    if (sel != null) {
        var saveInfo = (SaveDialogInfo)sel;
        var slot = (SlotInfo) saveInfo.SelectedItem;
        CompositionSaver.Save(slot.ExtensionInfo, new SlotInfo[] { slot }, saveInfo.TextDescription);
    }
  }

  private static int CompareSlotInfoAlphabetical(SlotInfo s1, SlotInfo s2)
  {
      if (s1.ExtensionInfo.Name.CompareTo(s2.ExtensionInfo.Name) == 0)
          return s1.Name.CompareTo(s2.Name);
      else
          return s1.ExtensionInfo.Name.CompareTo(s2.ExtensionInfo.Name);
  }

  internal static object ChooseDialog(Form parent, IEnumerable elements, string caption, SearchControl.FormatDelegate format) {
    var f = new SearchForm(elements) {
        StartPosition = FormStartPosition.CenterParent,
        Text = caption,
        Format = format
    };
    return (f.ShowDialog(parent) == DialogResult.OK) ? f.SelectedItem : null;
  }

  internal static object RestoreDialog(Form parent, IEnumerable elements, string caption, SearchControl.FormatDelegate format)
  {
      var f = new RestoreForm(elements)
      {
          StartPosition = FormStartPosition.CenterParent,
          Text = caption,
          Format = format
      };
      return (f.ShowDialog(parent) == DialogResult.OK) ? f.SelectedItem : null;
  }

  internal static object SaveDialog(Form parent, IEnumerable elements, string caption, SearchControl.FormatDelegate format) {
    var f = new SaveForm(elements) {
        StartPosition = FormStartPosition.CenterParent,
        Text = caption,
        Format = format
    };
    if (f.ShowDialog(parent) == DialogResult.OK)
    {
        return new SaveDialogInfo(f.SelectedItem, f.TextDescription);
    }
    else
        return null;
  }

  private class SaveDialogInfo
  {
      private object selectedItem;
      private string textDescription;

      public SaveDialogInfo(object sel, string desc)
      {
          this.selectedItem = sel;
          this.textDescription = desc;
      }

      public object SelectedItem { get { return this.selectedItem; } }
      public string TextDescription { get { return this.textDescription; } }
  }

  internal static List<PlugTypeInfo> GetRegisteredPlugs() {
    var plugs = new List<PlugTypeInfo>();
    foreach (var plug in Runtime.Repository.PlugTypeInfos) {
      if (plug.RegisteredInSlots.Count == 0) {
        continue;
      }
      plugs.Add(plug);
    }
    return plugs;
  }

  internal static List<PlugInfo> GetPluggedPlugs() {
    var plugs = new List<PlugInfo>();
    foreach (var plug in Runtime.Repository.PlugInfos) {
      if (plug.PluggedInSlots.Count == 0) {
        continue;
      }
      plugs.Add(plug);
    }
    return plugs;
  }

  internal static List<CompositionRestorer.SavedCompilationInfo> GetSavedCompilations() {
    return CompositionRestorer.GetSavedCompilations();
  }

  private static string FormatRegister(object obj) {
    var plug = (PlugTypeInfo) obj;
    return String.Format("{0} >> {1}", plug.ExtensionTypeInfo.Name,
        plug.Name.Substring(plug.Name.StartsWith("Plux.") ? 5 : 0));
  }

  private static string FormatPlug(object obj) {
    var plug = (PlugTypeInfo) obj;
    return String.Format("{0} >> {1}", plug.ExtensionTypeInfo.Name,
        plug.Name.Substring(plug.Name.StartsWith("Plux.") ? 5 : 0));
  }

  private static string FormatUnplug(object obj) {
    var plug = (PlugInfo) obj;
    return String.Format("{0} >> {1}", plug.ExtensionInfo.Name,
        plug.Name.Substring(plug.Name.StartsWith("Plux.") ? 5 : 0));
  }

  private static string FormatDeregister(object obj) {
    var plug = (PlugTypeInfo) obj;
    return String.Format("{0} << {1}", plug.ExtensionTypeInfo.Name,
        plug.Name.Substring(plug.Name.StartsWith("Plux.") ? 5 : 0));
  }

  private static string FormatRestore(object obj) {
      var storedComp = (CompositionRestorer.SavedCompilationInfo)obj;
      if(storedComp.Root == null)
        return String.Format("{0} !", storedComp.Description);
      else
          return String.Format("{0} - root: {1} ({2})", storedComp.Description, storedComp.Root.Name, storedComp.Root.Id);
  }

  private static string FormatSlot(object obj) {
    var slot = (SlotInfo) obj;
    return String.Format("{0} - {1} ({2})", slot.ExtensionInfo.Name, 
            slot.Name.Substring(slot.Name.StartsWith("Plux.") ? 5 : 0), slot.Id);
  }
}
